import os, sys
import pip

from emucorebrain.consts import consts
import importlib
import json
import numpy as np
import emucorebrain.keywords.model as keywords_model
from emucorebrain.core.preprocessor import PreProcessor
from emucorebrain.core.trainer import Trainer
from emucorebrain.data.models.route_model import RouteModel


class Predictor:

    ARG_KEY_MODEL_FILEPATH = "model_filepath"
    ARG_KEY_PREDICTION_THRESHOLD = "prediction_threshold"

    ARG_KEY_TRAINED_PREPROCESSOR = "trained_preprocessor"
    ARG_KEY_SYNAPSE0 = "synapse0"
    ARG_KEY_SYNAPSE1 = "synapse1"

    def __init__(self, *args, **dict_args):
        if len(dict_args) == 2:
            model_filepath = dict_args.get(self.ARG_KEY_MODEL_FILEPATH)
            prediction_threshold = dict_args.get(self.ARG_KEY_PREDICTION_THRESHOLD)

            with open(model_filepath, "r") as model_file:
                model_json = json.load(model_file)

                self._classes = model_json[keywords_model.save_file_classes]
                self._word_list = model_json[keywords_model.save_file_words]
                self._synapse0 = np.asarray(model_json[keywords_model.save_file_synapse0])
                self._synapse1 = np.asarray(model_json[keywords_model.save_file_synapse1])

            self._prediction_threshold = prediction_threshold

        elif len(dict_args) == 4:
            trained_preprocessor : PreProcessor = dict_args.get(self.ARG_KEY_TRAINED_PREPROCESSOR)
            synapse0 = dict_args.get(self.ARG_KEY_SYNAPSE0)
            synapse1 = dict_args.get(self.ARG_KEY_SYNAPSE1)
            prediction_threshold = dict_args.get(self.ARG_KEY_PREDICTION_THRESHOLD)

            self._classes = trained_preprocessor.get_unique_text_labels()
            self._word_list = trained_preprocessor.get_unique_wordlist()
            self._synapse0 = synapse0
            self._synapse1 = synapse1
            self._prediction_threshold = prediction_threshold

    def set_prediction_threshold(self, prediction_threshold):
        self._prediction_threshold = prediction_threshold

    def get_probable_classes_sentence(self, sentence):
        patternized_sentence = PreProcessor.get_sentence_patterns(sentence.lower(), self._word_list)

        # Input layer
        layer0 = patternized_sentence
        # Layer 1
        layer1 = Trainer.sigmoid(np.dot(layer0, self._synapse0))
        # Output layer
        layer2 = Trainer.sigmoid(np.dot(layer1, self._synapse1))

        results = [[i, probability] for i, probability in enumerate(layer2) if probability > self._prediction_threshold]
        results.sort(key=lambda x: x[1], reverse=True)
        return_results = [[self._classes[result[0]], result[1]] for result in results]

        return return_results

    def predict_sentence(self, sentence):
        probabilities = self.get_probable_classes_sentence(sentence)

        max_class = ""
        max_prob = 0

        for probability in probabilities:
            prob_class = probability[0]
            prob = probability[1]

            if prob > max_prob:
                max_class = prob_class
                max_prob = prob

        if max_prob > self._prediction_threshold:
            return max_class
        else:
            return None

    def get_unique_classes(self):
        return self._classes

    @staticmethod
    def _is_task_executor(class_object):
        try:
            return class_object.VALID_TASK_EXECUTOR

        except AttributeError:
            return False

    @staticmethod
    def install_dependencies(list_deps : list):
        for dep in list_deps:
            args = ["-q"]
            pip.main(["install"] + args + [dep])

    def run_validation_on_namespace_dir(self, tasks_namespaces_folderpath):
        structure_definitions_filepath = tasks_namespaces_folderpath + "/" + consts.TASKS_STRUCT_FILE_FILENAME
        if os.path.exists(structure_definitions_filepath):
            dep_filepath = tasks_namespaces_folderpath + "/" + consts.TASKS_DEPS_FILE_FILENAME
            if os.path.exists(dep_filepath):
                dep_file = open(dep_filepath, "rb")
                dep_data = json.load(dep_file)

                list_dep_data = []
                for dep_index in dep_data:
                    dep_index_data = dep_data[dep_index]

                    if consts.TASKS_DEPS_FILE_PROP_DEP_NAME in dep_index_data and consts.TASKS_DEPS_FILE_PROP_DEP_VERSION in dep_index_data:
                        dep_index_name = dep_index_data[consts.TASKS_DEPS_FILE_PROP_DEP_NAME]
                        dep_index_version = dep_index_data[consts.TASKS_DEPS_FILE_PROP_DEP_VERSION]
                        dep_index_req_line = dep_index_name + "==" + dep_index_version

                        list_dep_data.append(dep_index_req_line)
                    else:
                        raise Exception("Dependencies file contains malformed(missing name/version) near the dependency index: " + dep_index)

                self.install_dependencies(list_dep_data)

                structure_definitions_file_data = open(structure_definitions_filepath).read()
                structure_definitions_file_data = json.loads(structure_definitions_file_data)

                if all(basic_prop in structure_definitions_file_data for basic_prop in consts.TASKS_STRUCT_FILE_BASIC_PROPERTY_KEYS):
                    sys.path.append(os.path.abspath(tasks_namespaces_folderpath))

                    executors = structure_definitions_file_data[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS]
                    for executor in executors:
                        executor_folder = executor[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_NAMESPACE]
                        executor_folder_path = tasks_namespaces_folderpath + "/" + executor_folder

                        if os.path.exists(executor_folder_path):
                            executor_route_model = RouteModel(executor[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_CLASS])
                            executor_name = executor_route_model.get_name_task_executor()
                            executor_file_name = executor_name + ".py"
                            executor_file_path = executor_folder_path + "/" + executor_file_name

                            if not os.path.exists(executor_file_path):
                                raise Exception("Class File: " + executor_name + " cannot be found inside the Namespace Folder: " + executor_folder + ".")
                            else:
                                # Each namespace(task) file should have its class with the same name as the file
                                try:
                                    sys.path.append(os.path.abspath(executor_folder_path))

                                    unique_class_import_name = executor_name + "." + executor_name
                                    unique_class_object = getattr(importlib.import_module(unique_class_import_name), executor_name)
                                    assert self._is_task_executor(unique_class_object)

                                except AssertionError:
                                    raise Exception("One of the namespaces in the tasks directory does not implement TaskExecutor.")

                                except Exception:
                                    raise Exception("Unexpected error. MAYBE the file(s) in the tasks directory you've provided contains ambiguous class names.")
                        else:
                            raise Exception("Namespace folder: " + executor_folder + " not found.")

                else:
                    raise Exception("One or more of the following required entries are not found in the " + consts.TASKS_STRUCT_FILE_FILENAME + ".\n" + consts.TASKS_STRUCT_FILE_PROP_DEP_DIRS + ", " + consts.TASKS_STRUCT_FILE_PROP_EXECUTORS)
            else:
                raise Exception("Namespace folder: " + tasks_namespaces_folderpath + " does not contain the dependencies folder: " + consts.TASKS_DEPS_FILE_FILENAME)
        else:
            raise Exception(consts.TASKS_STRUCT_FILE_FILENAME + " does not exist inside " + tasks_namespaces_folderpath +".\nPlease make sure the tasks executors folder path you've given contains valid TaskExecutors.")

    def get_loaded_namespaces(self, tasks_namespaces_folderpath):
        # If the validation occurs successfully, we get the sys.path appended with all the namespace directories and
        # the parent directory. Therefore it in not necessary to append them here again.
        self.run_validation_on_namespace_dir(tasks_namespaces_folderpath)

        class_namespaces = {}

        structure_definitions_filepath = tasks_namespaces_folderpath + "/" + consts.TASKS_STRUCT_FILE_FILENAME
        structure_definitions_file_data = open(structure_definitions_filepath).read()
        structure_definitions_file_data = json.loads(structure_definitions_file_data)
        executors = structure_definitions_file_data[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS]
        for executor in executors:
            executor_route_model = RouteModel(executor[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_CLASS])
            executor_name = executor_route_model.get_name_task_executor()

            unique_class_import_name = executor_name + "." + executor_name
            class_namespace_class = getattr(importlib.import_module(unique_class_import_name), executor_name)
            class_namespace_instance = class_namespace_class()
            class_namespaces[executor_name] = class_namespace_instance

        return class_namespaces



import abc


# This is the "interface" class for all the input mechanisms
# Each an every input mechanism defined inside the mechanisms, should extend this class
import copy


class InputMechanism(abc.ABC):

    CONTAINER_KEY = None

    # This method is used to start listening on the device associated with the output mechanism
    # It would just be a simple function call to the input device which starts grabbing data from that respective
    # device.
    @abc.abstractmethod
    def start_listening(self):
        pass

    # This method is used to stop listening on the device associated with the output mechanism
    # It would just be a simple function call to the input device which stops grabbing data from that respective device.
    @abc.abstractmethod
    def stop_listening(self, raise_exception_if_stopped=True):
        pass

    # This method is used to read data from the input mechanism.
    # It may or may not block the thread, really depends on what mechanism is used.
    # To find out whether the thread is blocked by this method, head over to respective mechanism.
    # To find out what type of return value is expected head over to the respective mechanism.
    @abc.abstractmethod
    def read_data(self):
        pass

    # This method is used to return the GrabberController instance associated with the InputMechanism.
    # If GrabberControllers are not adopted in the implementation, this method may return None.
    # Otherwise, must return the associated GrabberController.
    @abc.abstractmethod
    def get_grabber_controller(self):
        pass


# Grabber is a wrapper around the function that is executed whenever an input is forwarded into it.
class Grabber:

    # Constructor for Grabber
    # on_exec: Function that should be executed when grabber receives data.
    def __init__(self, on_exec):
        self._on_exec = on_exec

    # Executes the function.
    # Accepts any number of positional or named arguments of any type which are directly passed into the function
    # executed.
    def exec(self, *args):
        self._on_exec(*args)


# GrabberController is a class that controls the Grabbers in a certain priority order.
class GrabberController:

    # The constant to identify the highest priority index.
    # Mostly used to debunk the confusion of index being least in numeric value over the highest priority.
    MAX_PRIORITY_INDEX = 0

    # Constructor for GrabberController
    # grabber_list: A python list which has it's items in the order of the priority.
    # notify_all: Set True to notify all the Grabbers in sequence of priority when data is received, False to notify
    # only the maximum priority Grabber.
    def __init__(self, grabber_list : list, notify_all=True):
        self._grabbers = {}
        for i, grabber in enumerate(grabber_list):
            self._grabbers[i] = grabber

        self._notify_all = notify_all

    # Adds a grabber into the grabber list prioritized by the given priority index.
    # If priority index is not followed up by the grabber, it is assumed as the least priority.
    # grabber: The new grabber to be added. An instance of Grabber class.
    # priority_index: The index of the priority that the grabber should be added into. Priority index is inversely
    # proportional to the priority. Higher the index, a lower priority is given. If left None, it is assumed that the
    # grabber is given the lowest priority available.
    def pop_in_grabber(self, grabber : Grabber, priority_index=None):
        # If priority index is None, assign it one more than the least priority of the current grabbers list,
        # i.e one more than the maximum index.
        if priority_index is None:
            priority_index = max(self._grabbers, key=int) + 1

        # Increases the lower priorities by 1 in order to add the new grabber at the specified priority index
        # In the case of the priority index being the least priority(highest index), re-assign the lower priorities
        # than the given priority a higher index, i.e. a more lower priority.
        lower_priority_grabber_indices = [key for key in self._grabbers if key >= priority_index]
        lower_priority_grabber_indices.sort(reverse=True)
        for lower_priority_index in lower_priority_grabber_indices:
            self._grabbers[lower_priority_index + 1] = self._grabbers[lower_priority_index]
            del self._grabbers[lower_priority_index]

        # Add the new grabber at the priority index.
        self._grabbers[priority_index] = grabber

    # Used to remove the Grabber at the given priority index.
    # priority_index: int representing the priority index at which the Grabber should be removed.
    # If left None, the Grabber with the least priority index(maximum index) is removed.
    def pop_out_grabber(self, priority_index=None):
        # If priority index is None, assign it the least priority of the current grabbers list, i.e the maximum index.
        if priority_index is None:
            priority_index = max(self._grabbers, key=int)

        # Removes the grabber at the specified priority index.
        if priority_index in self._grabbers:
            del self._grabbers[priority_index]
        else:
            raise Exception("There is no such priority index.")

        # Decreases the lower priorities by 1 as we have removed a grabber at a certain priority index.
        # In the case of the priority index being the least priority(highest index), re-assign the lower priorities
        # than the given priority a lower index, i.e. a more higher priority.
        lower_priority_grabber_indices = [key for key in self._grabbers if key > priority_index]
        lower_priority_grabber_indices.sort()
        for lower_priority_index in lower_priority_grabber_indices:
            self._grabbers[lower_priority_index - 1] = self._grabbers[lower_priority_index]
            del self._grabbers[lower_priority_index]

    # Used to get the Grabber at the given priority index.
    # priority_index: int representing the priority index at which the Grabber should be fetched.
    # If left None, the Grabber with the least priority index(maximum index) is returned.
    def get_out_grabber(self, priority_index=None):
        # If priority index is None, assign it the least priority of the current grabbers list, i.e the maximum index.
        if priority_index is None:
            priority_index = max(self._grabbers, key=int)

        return self._grabbers[priority_index]

    # Sets whether notify_grabbers method calls all the Grabbers' exec method in the order of their priority.
    # state: Set True to notify all the Grabbers in sequence of priority when data is received, False to notify
    # only the maximum priority Grabber.
    def set_notify_all(self, state : bool):
        self._notify_all = state

    # Executes the exec method of all the Grabbers in the order of maximum priority.
    # Pass any number of arguments to be passed directly to Grabber's exec method. Either positional or named.
    def notify_grabbers(self, *args):
        if self._notify_all:
            grabbers = copy.deepcopy(self._grabbers)
            for grabber_index in grabbers:
                grabber : Grabber = grabbers[grabber_index]
                grabber.exec(*args)
        else:
            if GrabberController.MAX_PRIORITY_INDEX in self._grabbers:
                grabber : Grabber = self._grabbers[GrabberController.MAX_PRIORITY_INDEX]
                grabber.exec(*args)
            else:
                raise Exception("There is no single Grabber to be executed.")



TASKS_STRUCT_FILE_FILENAME = "struct.json"
TASKS_STRUCT_FILE_PROP_DEP_DIRS = "dependency_directories"
TASKS_STRUCT_FILE_PROP_EXECUTORS = "executors"
TASKS_STRUCT_FILE_PROP_EXECUTORS_NAMESPACE = "namespace"
TASKS_STRUCT_FILE_PROP_EXECUTORS_CLASS = "class"
TASKS_STRUCT_FILE_PROP_EXECUTORS_PARAMS = "parameters"

TASKS_STRUCT_FILE_BASIC_PROPERTY_KEYS = [
    TASKS_STRUCT_FILE_PROP_DEP_DIRS,
    TASKS_STRUCT_FILE_PROP_EXECUTORS,
]

TASKS_DEPS_FILE_FILENAME = "dependencies.json"
TASKS_DEPS_FILE_PROP_DEP_NAME = "name"
TASKS_DEPS_FILE_PROP_DEP_VERSION = "version"



import importlib
from emucorebrain.data.abstracts.TaskExecutor import TaskExecutor


class RouteModel:

    INDEX_EXECUTOR = 0
    INDEX_METHOD = 1

    METHOD_SEPARATOR = '.'

    def __init__(self, route_str):
        self._route_str : str = route_str

    def get_name_task_executor(self):
        route_sep = self._route_str.split(self.METHOD_SEPARATOR)
        return route_sep[self.INDEX_EXECUTOR]

    # Please note that all the paths to the TaskExecutors namespaces should be added prior to calling this method.
    def get_task_executor(self):
        executor_name = self.get_name_executor_method()
        unique_class_import_name = executor_name + "." + executor_name
        class_namespace_class = getattr(importlib.import_module(unique_class_import_name), executor_name)
        class_namespace_instance = class_namespace_class()
        return class_namespace_instance

    def get_name_executor_method(self):
        route_sep = self._route_str.split(self.METHOD_SEPARATOR)
        if len(route_sep) > self.INDEX_METHOD:
            return route_sep[self.INDEX_METHOD]
        else:
            return TaskExecutor.DEFAULT_NAME_RUN_METHOD

    def get_executor_method(self):
        class_namespace_instance = self.get_task_executor()
        name_executor_method = self.get_name_executor_method()
        return getattr(class_namespace_instance, name_executor_method)

    def get_executor_method_by_instance(self, executor_instance : TaskExecutor):
        name_executor_method = self.get_name_executor_method()
        return getattr(executor_instance, name_executor_method)



from emucorebrain.consts import consts
import datetime
import json
import time
import numpy as np
from emucorebrain import keywords


class Trainer:

    def __init__(self, trained_preprocessor, tasks_namespaces_folderpath, train_verbose=False):
        self._trained_preprocessor = trained_preprocessor
        self._tasks_namespaces_folderpath = tasks_namespaces_folderpath
        self._train_verbose = train_verbose

        self._training_start_time = None
        self._training_end_time = None

        self._list_features = self._trained_preprocessor.get_processed_features()
        self._list_labels = self._trained_preprocessor.get_processed_labels()

        self._labels = np.array(self._list_labels)

        self._ptr_next_batch_begin = 0

        self._synapse_0 = None
        self._synapse_1 = None

    def _log(self, msg):
        if self._train_verbose:
            print(msg)

    # Compute sigmoid Non Linearity
    @staticmethod
    def sigmoid(x):
        output = 1 / (1 + np.exp(-x))
        return output

    # Convert output of sigmoid function to its derivative
    @staticmethod
    def sigmoid_output_to_derivative(output):
        return output * (1 - output)

    def _next_feature_batch(self, batch_size):
        next_batch_end = self._ptr_next_batch_begin + batch_size
        if next_batch_end > len(self._list_features):
            return_features_batch = self._list_features[self._ptr_next_batch_begin:len(self._list_features)]
            return_labels_batch = self._list_labels[self._ptr_next_batch_begin:len(self._list_features)]

            count_from_beginning = next_batch_end - len(self._list_features)

            feature_batch_from_beginning = self._list_features[0:count_from_beginning]
            return_features_batch.extend(feature_batch_from_beginning)

            labels_batch_from_beginning = self._list_labels[0:count_from_beginning]
            return_labels_batch.extend(labels_batch_from_beginning)

            self._ptr_next_batch_begin = count_from_beginning
        else:
            return_features_batch = self._list_features[self._ptr_next_batch_begin:next_batch_end]
            return_labels_batch = self._list_labels[self._ptr_next_batch_begin:next_batch_end]

            # Checking whether all features are obtained
            if next_batch_end > (len(self._list_features) - 1):
                next_batch_end = 0

            self._ptr_next_batch_begin = next_batch_end

        return return_features_batch, return_labels_batch

    def train(self, hidden_neurons=10, alpha=1, epochs=100, batch_size=None, iterations=100, dropout=False, dropout_percent=0.5):
        validation_state, _ = self._trained_preprocessor.validate_tasks_directory(self._tasks_namespaces_folderpath)
        if not validation_state:
            raise Exception("Tasks directory is invalid or does not contain all the namespaces.")

        self._training_start_time = time.time()
        self._training_end_time = None

        self._ptr_next_batch_begin = 0

        self._synapse_0 = None
        self._synapse_1 = None

        all_classes = self._trained_preprocessor.get_unique_text_labels()

        if batch_size is None:
            batch_size = len(self._list_features)
            iterations = 1

        self._log("Training with %s neurons, alpha:%s, dropout:%s %s" % (hidden_neurons, str(alpha), dropout, dropout_percent if dropout else ''))
        self._log("Input matrix: " + str(len(self._list_features)) + "x" + str(len(self._list_features[0])) + "     Output matrix: 1x" + str(len(all_classes)))
        np.random.seed(1)

        last_mean_error = 1
        # randomly initialize our weights with mean 0
        synapse_0 = 2 * np.random.random((len(self._list_features[0]), hidden_neurons)) - 1
        synapse_1 = 2 * np.random.random((hidden_neurons, len(all_classes))) - 1

        prev_synapse_0_weight_update = np.zeros_like(synapse_0)
        prev_synapse_1_weight_update = np.zeros_like(synapse_1)

        synapse_0_direction_count = np.zeros_like(synapse_0)
        synapse_1_direction_count = np.zeros_like(synapse_1)

        stop_epochs = False
        for j in iter(range(epochs + 1)):
            if stop_epochs:
                break
            else:
                for i in range(iterations):
                    nnp_feature_batch, nnp_labels_batch = self._next_feature_batch(batch_size)
                    feature_batch = np.array(nnp_feature_batch)
                    labels_batch = np.array(nnp_labels_batch)
                    # Feed forward through layers 0, 1, and 2
                    layer_0 = feature_batch
                    layer_1 = self.sigmoid(np.dot(layer_0, synapse_0))

                    if dropout:
                        layer_1 *= np.random.binomial([np.ones((len(feature_batch), hidden_neurons))], 1 - dropout_percent)[0] * (1.0 / (1 - dropout_percent))

                    layer_2 = self.sigmoid(np.dot(layer_1, synapse_1))

                    # how much did we miss the target value?
                    layer_2_error = labels_batch - layer_2

                    # Check this
                    if (j % 10000) == 0 and j > 5000:
                        # if this 10k epoch's error is greater than the last epoch, break out
                        if np.mean(np.abs(layer_2_error)) < last_mean_error:
                            self._log("delta after " + str(j) + " iterations:" + str(np.mean(np.abs(layer_2_error))))
                            last_mean_error = np.mean(np.abs(layer_2_error))
                        else:
                            self._log("break: " + str(np.mean(np.abs(layer_2_error))) + ">" + last_mean_error)
                            stop_epochs = True
                            break

                    # in what direction is the target value?
                    # were we really sure? if so, don't change too much.
                    layer_2_delta = layer_2_error * self.sigmoid_output_to_derivative(layer_2)

                    # how much did each l1 value contribute to the l2 error (according to the weights)?
                    layer_1_error = layer_2_delta.dot(synapse_1.T)

                    # in what direction is the target l1?
                    # were we really sure? if so, don't change too much.
                    layer_1_delta = layer_1_error * self.sigmoid_output_to_derivative(layer_1)

                    synapse_1_weight_update = (layer_1.T.dot(layer_2_delta))
                    synapse_0_weight_update = (layer_0.T.dot(layer_1_delta))

                    if j > 0:
                        synapse_0_direction_count += np.abs(((synapse_0_weight_update > 0) + 0) - ((prev_synapse_0_weight_update > 0) + 0))
                        synapse_1_direction_count += np.abs(((synapse_1_weight_update > 0) + 0) - ((prev_synapse_1_weight_update > 0) + 0))

                    synapse_1 += alpha * synapse_1_weight_update
                    synapse_0 += alpha * synapse_0_weight_update

                    prev_synapse_0_weight_update = synapse_0_weight_update
                    prev_synapse_1_weight_update = synapse_1_weight_update

        self._synapse_0 = synapse_0
        self._synapse_1 = synapse_1

        self._training_end_time = time.time()

    def save_model(self, savefile_path):
        now = datetime.datetime.now()

        words = self._trained_preprocessor.get_unique_wordlist()
        unique_classes = self._trained_preprocessor.get_unique_text_labels()

        synapse = {keywords.model.save_file_synapse0: self._synapse_0.tolist(), keywords.model.save_file_synapse1: self._synapse_1.tolist(),
                   keywords.model.save_file_datetime: now.strftime("%Y-%m-%d %H:%M"),
                   keywords.model.save_file_words: words,
                   keywords.model.save_file_classes: unique_classes
                   }

        with open(savefile_path, 'w') as outfile:
            json.dump(synapse, outfile, indent=4, sort_keys=True)

    def get_training_time(self):
        return self._training_end_time - self._training_start_time



from emucorebrain.data.carriers.carrier import Carrier


class StringCarrier(Carrier):

    CARRIER_TYPE = 2

    def __init__(self, data : str):
        self._data = data

    def get_data(self):
        return self._data

    def set_data(self, data : str):
        self._data = data



# This is the "interface" class for all the output mechanisms
# Each an every output mechanism defined inside the mechanisms, should extend this class
import abc


class OutputMechanism(abc.ABC):

    CONTAINER_KEY = None

    # This method is used to output/write data to the output mechanism.
    # It may or may not block the thread, really depends on what mechanism is used.
    # To find out whether the thread is blocked by this method, head over to respective mechanism.
    # To find out what type of return value is expected head over to the respective mechanism.
    @abc.abstractmethod
    def write_data(self, data, wait_until_completed=False):
        pass

    # This method is called to output any queued data if there are any related to any implemented output mechanism
    # Head over to the implementation for more explanation.
    @abc.abstractmethod
    def run_queued_data(self):
        pass




ARG_SPEECH_TEXT_DATA = "speech_data"
ARG_SETTINGS_CONTAINER = "settings"

ARG_OUTS_MECHANISMS_CARRIERS = "outs_mechanisms"
ARG_OUTS_MECHANISMS_MECHANISM_DEFAULT = "outs_mechanism_default"

ARG_INS_MECHANISMS_CARRIERS = "ins_mechanisms"
ARG_INS_MECHANISMS_MECHANISM_DEFAULT = "ins_mechanism_default"



import os
import json


class SettingsContainer:

    SETTING_DISPLAY_KEY = "display_key"
    SETTING_VALUE = "value"

    def __init__(self, file_path):
        self._file_path = file_path

        if not os.path.exists(file_path):
            settings_file_creator_file = open(file_path, "a+")
            json.dump({}, settings_file_creator_file)
            settings_file_creator_file.close()

        settings_file = open(self._file_path, "r+")
        self._settings = json.load(settings_file)
        settings_file.close()

    def get_setting_display_key(self, setting):
        return self._settings[setting][SettingsContainer.SETTING_DISPLAY_KEY]

    def get_setting(self, setting):
        return self._settings[setting][SettingsContainer.SETTING_VALUE]

    def set_setting(self, setting, value):
        self._settings[setting][SettingsContainer.SETTING_VALUE] = value

        settings_file = open(self._file_path, "r+")
        json.dump(self._settings, settings_file)
        settings_file.close()

    @staticmethod
    def is_valid_file(file_path):
        try:
            settings_file = open(file_path, "r+")
            json.load(settings_file)
            return True

        except:
            return False




# This is the "interface" class for all the tasks(basic functions that vehicle should perform)
# Each an every task should implement this.
import abc


class TaskExecutor(abc.ABC):

    VALID_TASK_EXECUTOR = True
    DEFAULT_NAME_RUN_METHOD = "run"

    # This method is used to handle any negative-type speech input to the same task type defined in run method.
    # This method is NOT the default method that would run if Task Executor is directly called.
    # Would have to explicitly call this method when defining training data.
    # args should be a dictionary of arguments that are taken inside set with the defined args inside the system.
    @abc.abstractmethod
    def run_negative(self, args):
        pass

    # The following method is directly called by the any other section of the program.
    # This is the function where one should implement the gist of the task to be performed.
    # No other methods defined by the implemented classes are called.
    # args should be a dictionary of arguments that are taken inside set with the defined args inside the system.
    @abc.abstractmethod
    def run(self, args):
        pass



from setuptools import setup

setup(name='emucorebrain',
      version='0.5.4',
      description='Brain of EmulationCore',
      url='https://github.com/GENIVI/GENIVI-GSoC-18/tree/master/Brain/core',
      author='Chandeepa Dissanayake',
      author_email='chandeepadissanayake@gmail.com',
      license='Mozilla Public License 2.0',
      packages=['emucorebrain.consts', 'emucorebrain.core', 'emucorebrain.data', 'emucorebrain.data.abstracts',
                'emucorebrain.data.carriers', 'emucorebrain.data.containers', 'emucorebrain.data.models',
                'emucorebrain.io', 'emucorebrain.io.mechanisms', 'emucorebrain.keywords'],
      install_requires=["nltk==3.3"],
      zip_safe=False
      )



from emucorebrain.consts import consts
import nltk
from nltk.stem.lancaster import LancasterStemmer
import json
import os


class PreProcessor:

    # nl_features : Sentences
    # text_labels : Python namespaces for the file to be mapped for respective sentence
    def __init__(self, nl_features, text_labels, stem_blacklist_words):
        self._nl_features = nl_features
        self._text_labels = text_labels
        self._stem_blacklist_words = stem_blacklist_words

        self._word_list = []
        self._unique_labels_list = []
        self._nl_feature_label_map = []

        self._stemmer = LancasterStemmer()

        self._features = []
        self._labels = []

    def tokenize_and_stem(self):
        for i in range(0, len(self._nl_features)):
            nl_feature = self._nl_features[i]
            text_label = self._text_labels[i]

            tokenized_feature = nltk.word_tokenize(nl_feature)
            self._word_list.extend(tokenized_feature)
            self._nl_feature_label_map.append((tokenized_feature, text_label))

            if text_label not in self._unique_labels_list:
                self._unique_labels_list.append(text_label)

        self._word_list = [self._stemmer.stem(word.lower()) for word in self._word_list if word not in self._stem_blacklist_words]
        self._word_list = list(set(self._word_list))

        self._unique_labels_list = list(set(self._unique_labels_list))

    def convert_to_patterns(self):
        encoded_label_template = [0] * len(self._unique_labels_list)

        for example in self._nl_feature_label_map:
            words_in_example = example[0] # Get the word list of the example
            label_of_example = example[1] # Get the label of the example

            words_in_example = [self._stemmer.stem(word_in_example.lower()) for word_in_example in words_in_example] # Stem each word

            pattern_for_words = []
            for unique_word in self._word_list:
                if unique_word in words_in_example:
                    pattern_for_words.append(1)
                else:
                    pattern_for_words.append(0)

            encoded_label = list(encoded_label_template)
            encoded_label[self._unique_labels_list.index(label_of_example)] = 1

            self._features.append(pattern_for_words)
            self._labels.append(encoded_label)

    def get_text_labels(self):
        return self._text_labels

    def get_unique_text_labels(self):
        return list(set(self._text_labels))

    def get_processed_features_and_labels(self):
        return self._features, self._labels

    def get_processed_features(self):
        return self._features

    def get_processed_labels(self):
        return self._labels

    def get_feature_length(self):
        return len(self._word_list)

    def get_unique_label_count(self):
        return len(self._unique_labels_list)

    def get_unique_wordlist(self):
        return self._word_list

    def validate_tasks_directory(self, tasks_directory_path):
        if os.path.exists(tasks_directory_path):
            structure_definitions_filepath = tasks_directory_path + "/" + consts.TASKS_STRUCT_FILE_FILENAME

            if os.path.exists(structure_definitions_filepath):
                structure_definitions_file_data = open(structure_definitions_filepath).read()
                structure_definitions_file_data = json.loads(structure_definitions_file_data)

                if all(basic_prop in structure_definitions_file_data for basic_prop in consts.TASKS_STRUCT_FILE_BASIC_PROPERTY_KEYS):

                    all_executors = structure_definitions_file_data[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS]

                    for executor in all_executors:
                        executor_folder = executor[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_NAMESPACE]
                        executor_folder_path = tasks_directory_path + "/" + executor_folder

                        if os.path.exists(executor_folder_path):
                            executor_name = executor[consts.TASKS_STRUCT_FILE_PROP_EXECUTORS_CLASS]
                            executor_file_name = executor_name + ".py"
                            executor_file_path = executor_folder_path + "/" + executor_file_name

                            if not os.path.exists(executor_file_path):
                                return False, "Class File: " + executor_name + " cannot be found inside the Namespace Folder: " + executor_folder + "."
                        else:
                            return False, "Namespace folder: " + executor_folder + " not found."

                    return True, "Success"
                else:
                    return False, "One or more of the following required entries are not found in the " + consts.TASKS_STRUCT_FILE_FILENAME + ".\n" + consts.TASKS_STRUCT_FILE_PROP_DEP_DIRS + ", " + consts.TASKS_STRUCT_FILE_PROP_EXECUTORS
            else:
                return False, consts.TASKS_STRUCT_FILE_FILENAME + " does not exist inside " + tasks_directory_path +".\nPlease make sure the tasks executors folder path you've given contains valid TaskExecutors."
        else:
            return False, tasks_directory_path + " is an invalid directory."

    def get_sentence_patterns(self, sentence):
        tokenized_words = nltk.word_tokenize(sentence)
        stemmed_words = [self._stemmer.stem(tokenized_word.lower()) for tokenized_word in tokenized_words]

        patterns = [0] * len(self._word_list)
        for stemmed_word in stemmed_words:
            for index, trainingSet_word in enumerate(self._word_list):
                if stemmed_word == trainingSet_word:
                    patterns[index] = 1

        return patterns

    @staticmethod
    def get_sentence_patterns(sentence, word_list):
        tokenized_words = nltk.word_tokenize(sentence)
        stemmer = LancasterStemmer()
        stemmed_words = [stemmer.stem(tokenized_word.lower()) for tokenized_word in tokenized_words]

        patterns = [0] * len(word_list)
        for stemmed_word in stemmed_words:
            for index, trainingSet_word in enumerate(word_list):
                if stemmed_word == trainingSet_word:
                    patterns[index] = 1

        return patterns
